package models

import (
	"ims/util/db/dts"
	"time"

	"gorm.io/gorm"
)

// ProductPower 产品权限
type ProductPower int8

const (
	SellPower     ProductPower = iota + 1 // 销售
	PurchasePower                         // 采购
	GiftPower                             // 赠送
)

func (p ProductPower) Valid() bool {
	return p >= SellPower && p <= GiftPower
}

// Product 产品
type Product struct {
	ID          uint                    `json:"id" gorm:"primarykey"`                    // 产品编号
	Code        string                  `json:"code" gorm:"uniqueIndex"`                 // 产品编码，为系统扩展和对外开放时使用
	Title       string                  `json:"title" gorm:"uniqueIndex"`                // 产品名称
	Description dts.NullString          `json:"description"`                             // 产品描述
	Content     dts.NullString          `json:"content" gorm:"type:text"`                // 产品内容
	CategoryID  uint                    `json:"category_id"`                             // 所属分类编号
	Category    *ProductCategory        `json:"category,omitempty"`                      // 所属分类
	Brand       dts.NullString          `json:"brand"`                                   // 品牌，可选
	Spec        dts.NullString          `json:"spec"`                                    // 产品规格，可选
	Unit        string                  `json:"unit"`                                    // 计量单位 - 台、套、箱、个、件、盒、片
	CostPrice   float64                 `json:"cost_price" gorm:"type:numeric(8,2)"`     // 成本单价，预设
	SalePrice   float64                 `json:"sale_price" gorm:"type:numeric(8,2)"`     // 标准售价（含税），预设
	Photos      []string                `json:"photos" gorm:"type:text;serializer:json"` // 产品相册
	Powers      dts.Slice[ProductPower] `json:"powers"`                                  // 产品权限
	Inbound     int                     `json:"inbound"`                                 // 总入库数量，历次入库累加的结果
	Outbound    int                     `json:"outbound"`                                // 总出库数量，历次出库累加的结果
	Stock       int                     `json:"stock"`                                   // 当前可用库存，与 Hold 相加之和为当前的总库存
	Hold        int                     `json:"hold"`                                    // 锁定（冻结）库存，与 Stock 相加之和为当前的总库存

	CreatedBy uint           `json:"created_by"`                             // 创建者（员工编号）
	CreatedAt time.Time      `json:"created_at"`                             // 创建时间
	UpdatedBy dts.NullUint   `json:"updated_by"`                             // 创建者（员工编号）
	UpdatedAt dts.NullTime   `json:"updated_at" gorm:"autoUpdateTime:false"` // 上次操作时间
	DeletedBy dts.NullUint   `json:"deleted_by"`                             // 删除者（员工编号）
	DeletedAt gorm.DeletedAt `json:"deleted_at" gorm:"index"`                // 删除数据时间

	Creator *Employee `json:"creator,omitempty" gorm:"foreignKey:CreatedBy"` // 创建数据的员工
	Updater *Employee `json:"updater,omitempty" gorm:"foreignKey:UpdatedBy"` // 上次更新数据的员工
	Deleter *Employee `json:"deleter,omitempty" gorm:"foreignKey:DeletedBy"` // 删除数据的员工
}
