package models

import (
	"ims/util/db/dts"
	"time"

	"gorm.io/gorm"
)

// TagType 标签类型
type TagType int8

const (
	TagCustomerType      TagType = iota + 1 // 客户分类，比如：潜在客户、成交客户、战略合作等；
	TagCustomerLabel                        // 客户标签
	TagPriceStrategy                        // 价格策略，比如：标准售价、一级售价、二级售价等；
	TagSettlementMethod                     // 结算方式，比如：现结、周结、月结、季结等
	TagSupplierType                         // 供应商分类
	TagSupplierLevel                        // 供应商等级
	TagWarehouseType                        // 仓库类型
	TagPurchaseReason                       // 需求原因，比如：补货、库存预警、销售需求等
	TagPaymentType                          // 支付方式，比如：银行转账、现金支付、微信支付、支付宝支付、银行支票等
	TagPaymentBatch                         // 支付批次，比如：整批货款、定金、第一批货款、第二批货款、第三批货款
	TagDistributionBatch                    // 配送批次，比如：整批、第一批、第二批、第三批等
	Tag_                                    // 在这个前面添加类型，而不用修改 Valid 方法
)

// Valid 是否有效
func (t TagType) Valid() bool {
	return t >= TagCustomerType && t < Tag_
}

// Tag 通用标签
type Tag struct {
	ID    uint    `json:"id"    gorm:"primarykey"`                 // 标签编号
	Type  TagType `json:"type"  gorm:"uniqueIndex:uni_typed_name"` // 标签类型
	Name  string  `json:"name"  gorm:"uniqueIndex:uni_typed_name"` // 标签名称
	Color string  `json:"color" gorm:"default:null"`               // 标签颜色
	Value dts.Any `json:"value"`                                   // 绑定的值
	Sort  int     `json:"sort"`                                    // 标签排序

	CreatedBy uint           `json:"created_by"`                             // 创建者（员工编号）
	CreatedAt time.Time      `json:"created_at"`                             // 创建时间
	UpdatedBy dts.NullUint   `json:"updated_by"`                             // 创建者（员工编号）
	UpdatedAt dts.NullTime   `json:"updated_at" gorm:"autoUpdateTime:false"` // 上次操作时间
	DeletedBy dts.NullUint   `json:"deleted_by"`                             // 删除者（员工编号）
	DeletedAt gorm.DeletedAt `json:"deleted_at" gorm:"index"`                // 删除数据时间

	Creator *Employee `json:"creator,omitempty" gorm:"foreignKey:CreatedBy"` // 创建数据的员工
	Updater *Employee `json:"updater,omitempty" gorm:"foreignKey:UpdatedBy"` // 上次更新数据的员工
	Deleter *Employee `json:"deleter,omitempty" gorm:"foreignKey:DeletedBy"` // 删除数据的员工
}
