package dts

import (
	"context"
	"database/sql/driver"
	"encoding/json"
	"errors"
	"fmt"
	"strings"

	"gorm.io/driver/mysql"
	"gorm.io/gorm"
	"gorm.io/gorm/clause"
	"gorm.io/gorm/schema"
)

// Any give a generic data type for json encoded data.
type Any struct {
	data any
}

func NewAny(data any) Any {
	return Any{
		data: data,
	}
}

// Data return data with generic Type T
func (j Any) Data() any {
	return j.data
}

// Value return json value, implement driver.Valuer interface
func (j Any) Value() (driver.Value, error) {
	return json.Marshal(j.data)
}

// Scan scan value into Any, implements sql.Scanner interface
func (j *Any) Scan(value any) error {
	var bytes []byte
	switch v := value.(type) {
	case []byte:
		bytes = v
	case string:
		bytes = []byte(v)
	default:
		return errors.New(fmt.Sprint("Failed to unmarshal JSONB value:", value))
	}
	return json.Unmarshal(bytes, &j.data)
}

// MarshalJSON to output non base64 encoded []byte
func (j Any) MarshalJSON() ([]byte, error) {
	return json.Marshal(j.data)
}

// UnmarshalJSON to deserialize []byte
func (j *Any) UnmarshalJSON(b []byte) error {
	return json.Unmarshal(b, &j.data)
}

// GormDataType gorm common data type
func (Any) GormDataType() string {
	return "json"
}

// GormDBDataType gorm db data type
func (Any) GormDBDataType(db *gorm.DB, field *schema.Field) string {
	switch db.Dialector.Name() {
	case "sqlite":
		return "JSON"
	case "mysql":
		return "JSON"
	case "postgres":
		return "JSONB"
	}
	return ""
}

func (js Any) GormValue(ctx context.Context, db *gorm.DB) clause.Expr {
	data, _ := js.MarshalJSON()

	switch db.Dialector.Name() {
	case "mysql":
		if v, ok := db.Dialector.(*mysql.Dialector); ok && !strings.Contains(v.ServerVersion, "MariaDB") {
			return gorm.Expr("CAST(? AS JSON)", string(data))
		}
	}

	return gorm.Expr("?", string(data))
}
