package dts

import (
	"database/sql"
	"database/sql/driver"
	"encoding/json"
	"time"
)

// NullTime represents a [time.Time] that may be null.
// NullTime implements the [Scanner] interface so
// it can be used as a scan destination, similar to [NullString].
type NullTime struct {
	Time  time.Time
	Valid bool // Valid is true if Time is not NULL
}

// Scan implements the Scanner interface.
func (nt *NullTime) Scan(value any) error {
	if value == nil {
		nt.Time, nt.Valid = time.Time{}, false
		return nil
	}
	st := new(sql.NullTime)
	err := st.Scan(value)
	nt.Time = st.Time
	nt.Valid = st.Valid
	return err
}

// Value implements the driver Valuer interface.
func (nt NullTime) Value() (driver.Value, error) {
	if !nt.Valid {
		return nil, nil
	}
	return nt.Time, nil
}

func (nt *NullTime) UnmarshalJSON(bytes []byte) error {
	err := nt.Time.UnmarshalJSON(bytes)
	if err != nil {
		return err
	}
	nt.Valid = !nt.Time.IsZero()
	return nil
}

func (nt NullTime) MarshalJSON() ([]byte, error) {
	if !nt.Valid {
		return json.Marshal(nil)
	}
	return nt.Time.MarshalJSON()
}
